@@ -12,12 +12,23 @@
 
 var ConnectionHandler = require('./connection/ConnectionHandler');
 
+var MutationHandlers = require('./connection/MutationHandlers');
+
 var invariant = require("fbjs/lib/invariant");
 
 function RelayDefaultHandlerProvider(handle) {
   switch (handle) {
     case 'connection':
       return ConnectionHandler;
+
+    case 'deleteRecord':
+      return MutationHandlers.DeleteRecordHandler;
+
+    case 'appendEdge':
+      return MutationHandlers.AppendEdgeHandler;
+
+    case 'prependEdge':
+      return MutationHandlers.PrependEdgeHandler;
   }
 
   !false ? process.env.NODE_ENV !== "production" ? invariant(false, 'RelayDefaultHandlerProvider: No handler provided for `%s`.', handle) : invariant(false) : void 0;
