@@ -14,6 +14,8 @@ var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefau
 
 var _toConsumableArray2 = _interopRequireDefault(require("@babel/runtime/helpers/toConsumableArray"));
 
+var RelayFeatureFlags = require('../util/RelayFeatureFlags');
+
 var RelayModernRecord = require('./RelayModernRecord');
 
 var RelayProfiler = require('../util/RelayProfiler');
@@ -32,7 +34,8 @@ var _require = require('../util/RelayConcreteNode'),
     MODULE_IMPORT = _require.MODULE_IMPORT,
     SCALAR_FIELD = _require.SCALAR_FIELD,
     SCALAR_HANDLE = _require.SCALAR_HANDLE,
-    STREAM = _require.STREAM;
+    STREAM = _require.STREAM,
+    TYPE_DISCRIMINATOR = _require.TYPE_DISCRIMINATOR;
 
 var _require2 = require('./ClientID'),
     generateClientID = _require2.generateClientID,
@@ -50,6 +53,10 @@ var _require4 = require('./RelayStoreUtils'),
     TYPENAME_KEY = _require4.TYPENAME_KEY,
     ROOT_ID = _require4.ROOT_ID;
 
+var _require5 = require('./TypeID'),
+    generateTypeID = _require5.generateTypeID,
+    TYPE_SCHEMA_TYPE = _require5.TYPE_SCHEMA_TYPE;
+
 /**
  * Normalizes the results of a query and standard GraphQL response, writing the
  * normalized records/fields into the given MutableRecordSource.
@@ -75,10 +82,10 @@ var RelayResponseNormalizer = /*#__PURE__*/function () {
     this._treatMissingFieldsAsNull = options.treatMissingFieldsAsNull;
     this._incrementalPlaceholders = [];
     this._isClientExtension = false;
+    this._isUnmatchedAbstractType = false;
     this._moduleImportPayloads = [];
     this._path = options.path ? (0, _toConsumableArray2["default"])(options.path) : [];
     this._recordSource = recordSource;
-    this._request = options.request;
     this._variables = variables;
   }
 
@@ -133,13 +140,76 @@ var RelayResponseNormalizer = /*#__PURE__*/function () {
           break;
 
         case INLINE_FRAGMENT:
-          var _typeName = RelayModernRecord.getType(record);
+          {
+            var abstractKey = selection.abstractKey;
 
-          if (_typeName === selection.type) {
-            this._traverseSelections(selection, record, data);
+            if (abstractKey == null) {
+              var _typeName = RelayModernRecord.getType(record);
+
+              if (_typeName === selection.type) {
+                this._traverseSelections(selection, record, data);
+              }
+            } else if (RelayFeatureFlags.ENABLE_PRECISE_TYPE_REFINEMENT) {
+              var implementsInterface = data.hasOwnProperty(abstractKey);
+
+              var _typeName2 = RelayModernRecord.getType(record);
+
+              var typeID = generateTypeID(_typeName2);
+
+              var typeRecord = this._recordSource.get(typeID);
+
+              if (typeRecord == null) {
+                typeRecord = RelayModernRecord.create(typeID, TYPE_SCHEMA_TYPE);
+
+                this._recordSource.set(typeID, typeRecord);
+              }
+
+              RelayModernRecord.setValue(typeRecord, abstractKey, implementsInterface);
+
+              if (implementsInterface) {
+                this._traverseSelections(selection, record, data);
+              }
+            } else {
+              // legacy behavior for abstract refinements: always normalize even
+              // if the type doesn't conform, but track if the type matches or not
+              // for determining whether response fields are expected to be present
+              var _implementsInterface = data.hasOwnProperty(abstractKey);
+
+              var parentIsUnmatchedAbstractType = this._isUnmatchedAbstractType;
+              this._isUnmatchedAbstractType = this._isUnmatchedAbstractType || !_implementsInterface;
+
+              this._traverseSelections(selection, record, data);
+
+              this._isUnmatchedAbstractType = parentIsUnmatchedAbstractType;
+            }
+
+            break;
           }
 
-          break;
+        case TYPE_DISCRIMINATOR:
+          {
+            if (RelayFeatureFlags.ENABLE_PRECISE_TYPE_REFINEMENT) {
+              var _abstractKey = selection.abstractKey;
+
+              var _implementsInterface2 = data.hasOwnProperty(_abstractKey);
+
+              var _typeName3 = RelayModernRecord.getType(record);
+
+              var _typeID = generateTypeID(_typeName3);
+
+              var _typeRecord = this._recordSource.get(_typeID);
+
+              if (_typeRecord == null) {
+                _typeRecord = RelayModernRecord.create(_typeID, TYPE_SCHEMA_TYPE);
+
+                this._recordSource.set(_typeID, _typeRecord);
+              }
+
+              RelayModernRecord.setValue(_typeRecord, _abstractKey, _implementsInterface2);
+            }
+
+            break;
+          }
 
         case LINKED_HANDLE:
         case SCALAR_HANDLE:
@@ -152,7 +222,8 @@ var RelayResponseNormalizer = /*#__PURE__*/function () {
             dataID: RelayModernRecord.getDataID(record),
             fieldKey: fieldKey,
             handle: selection.handle,
-            handleKey: handleKey
+            handleKey: handleKey,
+            handleArgs: selection.handleArgs ? getArgumentValues(selection.handleArgs, this._variables) : {}
           });
 
           break;
@@ -268,21 +339,35 @@ var RelayResponseNormalizer = /*#__PURE__*/function () {
     var fieldValue = data[responseKey];
 
     if (fieldValue == null) {
-      if (!this._treatMissingFieldsAsNull && fieldValue === undefined) {
-        // Fields that are missing in the response are not set on the record.
-        // There are three main cases where this can occur:
+      if (fieldValue === undefined) {
+        // Fields may be missing in the response in two main cases:
         // - Inside a client extension: the server will not generally return
         //   values for these fields, but a local update may provide them.
-        // - Fields on abstract types: these may be missing if the concrete
-        //   response type does not match the abstract type.
-        //
-        // Otherwise, missing fields usually indicate a server or user error (
-        // the latter for manually constructed payloads).
-        if (process.env.NODE_ENV !== "production") {
-          process.env.NODE_ENV !== "production" ? warning(this._isClientExtension || parent.kind === LINKED_FIELD && parent.concreteType == null ? true : Object.prototype.hasOwnProperty.call(data, responseKey), 'RelayResponseNormalizer: Payload did not contain a value ' + 'for field `%s: %s`. Check that you are parsing with the same ' + 'query that was used to fetch the payload.', responseKey, storageKey) : void 0;
-        }
+        // - Inside an abstract type refinement where the concrete type does
+        //   not conform to the interface/union.
+        // However an otherwise-required field may also be missing if the server
+        // is configured to skip fields with `null` values, in which case the
+        // client is assumed to be correctly configured with
+        // treatMissingFieldsAsNull=true.
+        var isOptionalField = this._isClientExtension || this._isUnmatchedAbstractType;
 
-        return;
+        if (isOptionalField) {
+          // Field not expected to exist regardless of whether the server is pruning null
+          // fields or not.
+          return;
+        } else if (!this._treatMissingFieldsAsNull) {
+          // Not optional and the server is not pruning null fields: field is expected
+          // to be present
+          if (process.env.NODE_ENV !== "production") {
+            process.env.NODE_ENV !== "production" ? warning(false, 'RelayResponseNormalizer: Payload did not contain a value ' + 'for field `%s: %s`. Check that you are parsing with the same ' + 'query that was used to fetch the payload.', responseKey, storageKey) : void 0;
+          }
+
+          return;
+        }
+      }
+
+      if (selection.kind === SCALAR_FIELD && process.env.NODE_ENV !== "production") {
+        this._validateConflictingFieldsWithIdenticalId(record, storageKey, fieldValue);
       }
 
       RelayModernRecord.setValue(record, storageKey, null);
@@ -290,6 +375,8 @@ var RelayResponseNormalizer = /*#__PURE__*/function () {
     }
 
     if (selection.kind === SCALAR_FIELD) {
+      this._validateConflictingFieldsWithIdenticalId(record, storageKey, fieldValue);
+
       RelayModernRecord.setValue(record, storageKey, fieldValue);
     } else if (selection.kind === LINKED_FIELD) {
       this._path.push(responseKey);
@@ -311,36 +398,30 @@ var RelayResponseNormalizer = /*#__PURE__*/function () {
     var _field$concreteType;
 
     !(typeof fieldValue === 'object' && fieldValue) ? process.env.NODE_ENV !== "production" ? invariant(false, 'RelayResponseNormalizer: Expected data for field `%s` to be an object.', storageKey) : invariant(false) : void 0;
-    var nextID = this._getDataId(
-    /* $FlowFixMe(>=0.98.0 site=www,mobile,react_native_fb,oss) This comment
-     * suppresses an error found when Flow v0.98 was deployed. To see the
-     * error delete this comment and run Flow. */
-    fieldValue,
-    /* $FlowFixMe(>=0.98.0 site=www,mobile,react_native_fb,oss) This comment
-     * suppresses an error found when Flow v0.98 was deployed. To see the
-     * error delete this comment and run Flow. */
+    var nextID = this._getDataId( // $FlowFixMe[incompatible-variance]
+    fieldValue, // $FlowFixMe[incompatible-variance]
     (_field$concreteType = field.concreteType) !== null && _field$concreteType !== void 0 ? _field$concreteType : this._getRecordType(fieldValue)) || // Reuse previously generated client IDs
     RelayModernRecord.getLinkedRecordID(record, storageKey) || generateClientID(RelayModernRecord.getDataID(record), storageKey);
     !(typeof nextID === 'string') ? process.env.NODE_ENV !== "production" ? invariant(false, 'RelayResponseNormalizer: Expected id on field `%s` to be a string.', storageKey) : invariant(false) : void 0;
+
+    if (process.env.NODE_ENV !== "production") {
+      this._validateConflictingLinkedFieldsWithIdenticalId(record, RelayModernRecord.getLinkedRecordID(record, storageKey), nextID, storageKey);
+    }
+
     RelayModernRecord.setLinkedRecordID(record, storageKey, nextID);
 
     var nextRecord = this._recordSource.get(nextID);
 
     if (!nextRecord) {
-      /* $FlowFixMe(>=0.98.0 site=www,mobile,react_native_fb,oss) This comment
-       * suppresses an error found when Flow v0.98 was deployed. To see the
-       * error delete this comment and run Flow. */
-      var _typeName2 = field.concreteType || this._getRecordType(fieldValue);
+      // $FlowFixMe[incompatible-variance]
+      var _typeName4 = field.concreteType || this._getRecordType(fieldValue);
 
-      nextRecord = RelayModernRecord.create(nextID, _typeName2);
+      nextRecord = RelayModernRecord.create(nextID, _typeName4);
 
       this._recordSource.set(nextID, nextRecord);
     } else if (process.env.NODE_ENV !== "production") {
       this._validateRecordType(nextRecord, field, fieldValue);
-    }
-    /* $FlowFixMe(>=0.98.0 site=www,mobile,react_native_fb,oss) This comment
-     * suppresses an error found when Flow v0.98 was deployed. To see the error
-     * delete this comment and run Flow. */
+    } // $FlowFixMe[incompatible-variance]
 
 
     this._traverseSelections(field, nextRecord, fieldValue);
@@ -364,14 +445,8 @@ var RelayResponseNormalizer = /*#__PURE__*/function () {
       _this._path.push(String(nextIndex));
 
       !(typeof item === 'object') ? process.env.NODE_ENV !== "production" ? invariant(false, 'RelayResponseNormalizer: Expected elements for field `%s` to be ' + 'objects.', storageKey) : invariant(false) : void 0;
-      var nextID = _this._getDataId(
-      /* $FlowFixMe(>=0.98.0 site=www,mobile,react_native_fb,oss) This comment
-       * suppresses an error found when Flow v0.98 was deployed. To see the
-       * error delete this comment and run Flow. */
-      item,
-      /* $FlowFixMe(>=0.98.0 site=www,mobile,react_native_fb,oss) This comment
-       * suppresses an error found when Flow v0.98 was deployed. To see the
-       * error delete this comment and run Flow. */
+      var nextID = _this._getDataId( // $FlowFixMe[incompatible-variance]
+      item, // $FlowFixMe[incompatible-variance]
       (_field$concreteType2 = field.concreteType) !== null && _field$concreteType2 !== void 0 ? _field$concreteType2 : _this._getRecordType(item)) || prevIDs && prevIDs[nextIndex] || // Reuse previously generated client IDs:
       generateClientID(RelayModernRecord.getDataID(record), storageKey, nextIndex);
       !(typeof nextID === 'string') ? process.env.NODE_ENV !== "production" ? invariant(false, 'RelayResponseNormalizer: Expected id of elements of field `%s` to ' + 'be strings.', storageKey) : invariant(false) : void 0;
@@ -380,20 +455,19 @@ var RelayResponseNormalizer = /*#__PURE__*/function () {
       var nextRecord = _this._recordSource.get(nextID);
 
       if (!nextRecord) {
-        /* $FlowFixMe(>=0.98.0 site=www,mobile,react_native_fb,oss) This comment
-         * suppresses an error found when Flow v0.98 was deployed. To see the
-         * error delete this comment and run Flow. */
-        var _typeName3 = field.concreteType || _this._getRecordType(item);
+        // $FlowFixMe[incompatible-variance]
+        var _typeName5 = field.concreteType || _this._getRecordType(item);
 
-        nextRecord = RelayModernRecord.create(nextID, _typeName3);
+        nextRecord = RelayModernRecord.create(nextID, _typeName5);
 
         _this._recordSource.set(nextID, nextRecord);
       } else if (process.env.NODE_ENV !== "production") {
         _this._validateRecordType(nextRecord, field, item);
       }
-      /* $FlowFixMe(>=0.98.0 site=www,mobile,react_native_fb,oss) This comment
-       * suppresses an error found when Flow v0.98 was deployed. To see the
-       * error delete this comment and run Flow. */
+
+      if (prevIDs && process.env.NODE_ENV !== "production") {
+        _this._validateConflictingLinkedFieldsWithIdenticalId(record, prevIDs[nextIndex], nextID, storageKey);
+      } // $FlowFixMe[incompatible-variance]
 
 
       _this._traverseSelections(field, nextRecord, item);
@@ -413,6 +487,28 @@ var RelayResponseNormalizer = /*#__PURE__*/function () {
     var typeName = (_field$concreteType3 = field.concreteType) !== null && _field$concreteType3 !== void 0 ? _field$concreteType3 : this._getRecordType(payload);
     var dataID = RelayModernRecord.getDataID(record);
     process.env.NODE_ENV !== "production" ? warning(isClientID(dataID) && dataID !== ROOT_ID || RelayModernRecord.getType(record) === typeName, 'RelayResponseNormalizer: Invalid record `%s`. Expected %s to be ' + 'consistent, but the record was assigned conflicting types `%s` ' + 'and `%s`. The GraphQL server likely violated the globally unique ' + 'id requirement by returning the same id for different objects.', dataID, TYPENAME_KEY, RelayModernRecord.getType(record), typeName) : void 0;
+  }
+  /**
+   * Warns if a single response contains conflicting fields with the same id
+   */
+  ;
+
+  _proto._validateConflictingFieldsWithIdenticalId = function _validateConflictingFieldsWithIdenticalId(record, storageKey, fieldValue) {
+    if (process.env.NODE_ENV !== "production") {
+      var dataID = RelayModernRecord.getDataID(record);
+      var previousValue = RelayModernRecord.getValue(record, storageKey);
+      process.env.NODE_ENV !== "production" ? warning(storageKey === TYPENAME_KEY || previousValue === undefined || previousValue === fieldValue, 'RelayResponseNormalizer: Invalid record. The record contains two ' + 'instances of the same id: `%s` with conflicting field, %s and its values: %s and %s. ' + 'If two fields are different but share ' + 'the same id, one field will overwrite the other.', dataID, storageKey, previousValue, fieldValue) : void 0;
+    }
+  }
+  /**
+   * Warns if a single response contains conflicting fields with the same id
+   */
+  ;
+
+  _proto._validateConflictingLinkedFieldsWithIdenticalId = function _validateConflictingLinkedFieldsWithIdenticalId(record, prevID, nextID, storageKey) {
+    if (process.env.NODE_ENV !== "production") {
+      process.env.NODE_ENV !== "production" ? warning(prevID === undefined || prevID === nextID, 'RelayResponseNormalizer: Invalid record. The record contains ' + 'references to the conflicting field, %s and its id values: %s and %s. ' + 'We need to make sure that the record the field points ' + 'to remains consistent or one field will overwrite the other.', storageKey, prevID, nextID) : void 0;
+    }
   };
 
   return RelayResponseNormalizer;
